from sql.connection import ConnectionManager
from IPython import get_ipython
import math
import time
from sql.util import parse_sql_results_to_json
from sql.inspect import fetch_sql_with_pagination, is_table_exists
from sql.widgets import utils
from sql.telemetry import telemetry

import os
from ploomber_core.dependencies import check_installed

# Widget base dir
BASE_DIR = os.path.dirname(__file__)


class TableWidget:
    @telemetry.log_call("TableWidget-init")
    def __init__(self, table, schema=None):
        """
        Creates an HTML table element and populates it with SQL table

        Parameters
        ----------
        table : str
            Table name where the data is located
        """

        self.html = ""

        is_table_exists(table, schema)

        # load css
        html_style = utils.load_css(f"{BASE_DIR}/css/tableWidget.css")
        self.add_to_html(html_style)

        self.create_table(table, schema)

        # register listener for jupyter lab
        self.register_comm()

        # load_tests
        self.load_tests()

    def _repr_html_(self):
        return self.html

    def add_to_html(self, html):
        self.html += html

    def create_table(self, table, schema):
        """
        Creates an HTML table with default data
        """
        if schema:
            table_ = f"{schema}.{table}"
        else:
            table_ = table

        rows_per_page = 10
        rows, columns = fetch_sql_with_pagination(table_, 0, rows_per_page)
        rows = parse_sql_results_to_json(rows, columns)

        query = f"SELECT count(*) FROM {table_}"
        n_total = ConnectionManager.current.raw_execute(query).fetchone()[0]
        table_name = table_.strip('"').strip("'")

        n_pages = math.ceil(n_total / rows_per_page)

        unique_id = str(int(time.time()))
        table_container_id = f"tableContainer_{unique_id}"

        # Create table container with unique id
        table_container_html = f"""
            <div id="{table_container_id}" class="table-container"></div>
            """
        self.add_to_html(table_container_html)

        html_scripts = utils.load_js(
            [
                f"{BASE_DIR}/js/tableWidget.js",
                utils.set_template_params(
                    columns=list(columns),
                    rows_per_page=rows_per_page,
                    n_pages=n_pages,
                    n_total=n_total,
                    table_name=table_name,
                    table_container_id=table_container_id,
                    table=table_,
                    initialRows=rows,
                ),
            ]
        )
        self.add_to_html(html_scripts)

    def load_tests(self):
        """
        Define which JS functions we should
        include in this widget's test unit.

        Example:

        Given following the html:

        <script>

            function drawList() {
                return `<ul><li>item1</li><li>item2</li></ul>`
            }

            function drawTable() {
                return `
                        <table>
                            <tr>
                                <td>1</td>
                                <td>2</td>
                            </tr>
                        </table>
                    `
            }

        </script>

        We can include `drawList` in the test unit by extracting it
        from the html and add it to this widget's test property.

        self.tests["drawList"] = utils.extract_function_by_name(
            html, "drawList"
        )


        Testing with pytest:

        import js2py

        def test_draw_list(expected):
            expected = "<ul><li>item1</li><li>item2</li></ul>"

            table_widget = TableWidget("empty_table")

            result = js2py.eval_js(table_widget.tests["drawList"])

            assert result == expected

        """
        self.tests = dict()
        self.tests["createTableRows"] = utils.extract_function_by_name(
            self.html, "createTableRows"
        )

    def register_comm(self):
        """
        Register communication between the frontend and the kernel.
        """

        check_installed(
            ["jupysql_plugin"], "jupysql-plugin", pip_names=["jupysql-plugin"]
        )

        def comm_handler(comm, open_msg):
            """
            Handle received messages from the frontend
            """

            @comm.on_msg
            def _recv(msg):
                data = msg["content"]["data"]
                n_rows = data["nRows"]
                page = data["page"]

                sort_column = None
                sort_order = None
                table_name = data["table"]

                if "sort" in data:
                    sort = data["sort"]
                    sort_column = sort["column"]
                    sort_order = sort["order"]

                offset = page * n_rows

                rows, columns = fetch_sql_with_pagination(
                    table_name,
                    offset,
                    n_rows,
                    sort_column=sort_column,
                    sort_order=sort_order,
                )
                rows_json = parse_sql_results_to_json(rows, columns)

                comm.send({"rows": rows_json})

        ipython = get_ipython()

        if hasattr(ipython, "kernel"):
            ipython.kernel.comm_manager.register_target(
                "comm_target_handle_table_widget", comm_handler
            )



import sys
import argparse
import shlex

from IPython.core.magic import Magics, line_magic, magics_class, no_var_expand
from IPython.core.magic_arguments import argument, magic_arguments
from sql.inspect import support_only_sql_alchemy_connection
from sql.cmd.tables import tables
from sql.cmd.columns import columns
from sql.cmd.test import test
from sql.cmd.profile import profile
from sql.cmd.explore import explore
from sql.cmd.snippets import snippets
from sql.cmd.connect import connect
from sql.connection import ConnectionManager
from sql.util import check_duplicate_arguments

try:
    from traitlets.config.configurable import Configurable
except ModuleNotFoundError:
    from IPython.config.configurable import Configurable
from sql import exceptions


class CmdParser(argparse.ArgumentParser):
    def exit(self, status=0, message=None):
        if message:
            self._print_message(message, sys.stderr)

    def error(self, message):
        raise exceptions.UsageError(message)


@magics_class
class SqlCmdMagic(Magics, Configurable):
    """%sqlcmd magic"""

    @no_var_expand
    @line_magic("sqlcmd")
    @magic_arguments()
    @argument("line", type=str, help="Command name")
    def _validate_execute_inputs(self, line):
        """
        Function to validate %sqlcmd inputs.
        Raises UsageError in case of an invalid input, executes command otherwise.
        """

        # We rely on SQLAlchemy when inspecting tables

        AVAILABLE_SQLCMD_COMMANDS = [
            "tables",
            "columns",
            "test",
            "profile",
            "explore",
            "snippets",
            "connect",
        ]
        COMMANDS_CONNECTION_REQUIRED = [
            "tables",
            "columns",
            "test",
            "profile",
            "explore",
        ]
        COMMANDS_SQLALCHEMY_ONLY = ["tables", "columns", "test", "explore"]

        VALID_COMMANDS_MSG = (
            f"Missing argument for %sqlcmd. "
            f"Valid commands are: {', '.join(AVAILABLE_SQLCMD_COMMANDS)}"
        )

        if line == "":
            raise exceptions.UsageError(VALID_COMMANDS_MSG)
        else:
            # directly use shlex since SqlCmdMagic does not use magic_args from parse.py
            split = shlex.split(line, posix=False)
            command, others = split[0].strip(), split[1:]
            if others:
                check_duplicate_arguments(
                    self.execute,
                    "sqlcmd",
                    split,
                    disallowed_aliases={
                        "-t": "--table",
                        "-s": "--schema",
                        "-o": "--output",
                    },
                )

            if command in AVAILABLE_SQLCMD_COMMANDS:
                if (
                    command in COMMANDS_CONNECTION_REQUIRED
                    and not ConnectionManager.current
                ):
                    raise exceptions.RuntimeError(
                        f"Cannot execute %sqlcmd {command} because there "
                        "is no active connection. Connect to a database "
                        "and try again."
                    )

                if command in COMMANDS_SQLALCHEMY_ONLY:
                    support_only_sql_alchemy_connection(f"%sqlcmd {command}")

                return self.execute(command, others)
            else:
                raise exceptions.UsageError(
                    f"%sqlcmd has no command: {command!r}. "
                    "Valid commands are: {}".format(
                        ", ".join(AVAILABLE_SQLCMD_COMMANDS)
                    )
                )

    @argument("cmd_name", default="", type=str, help="Command name")
    @argument("others", default="", type=str, help="Other tags")
    def execute(self, cmd_name="", others="", cell="", local_ns=None):
        """
        Command
        """

        router = {
            "tables": tables,
            "columns": columns,
            "test": test,
            "profile": profile,
            "explore": explore,
            "snippets": snippets,
            "connect": connect,
        }

        cmd = router.get(cmd_name)
        if cmd_name == "connect":
            return cmd(others)
        else:
            return cmd(others, self.shell.user_ns.copy())



"""Get/set the current SqlMagic instance."""

__sql_magic = None


def _get_sql_magic():
    """Returns the current SqlMagic instance."""
    if __sql_magic is None:
        raise RuntimeError("%sql has not been loaded yet. Run %load_ext sql")

    return __sql_magic


def _set_sql_magic(sql_magic):
    """Sets the current SqlMagic instance."""
    global __sql_magic
    __sql_magic = sql_magic


def _config_feedback_all():
    """Returns True if the current feedback level is >=2"""
    return _get_sql_magic().feedback >= 2


def _config_feedback_normal_or_more():
    """Returns True if the current feedback level is >=1"""
    return _get_sql_magic().feedback >= 1



import re
import itertools
import shlex
from os.path import expandvars
from pathlib import Path
import configparser
import warnings

from sqlalchemy.engine.url import URL

from sql import exceptions
from sql.util import check_duplicate_arguments

# Keywords used to identify the beginning of SQL queries
# in split_args_and_sql(). Should cover all cases but can
# be edited to include special keywords.
SQL_COMMANDS = [
    "select",
    "from",
    "with",
    "pivot",
    "create",
    "update",
    "delete",
    "insert",
    "alter",
    "drop",
    "describe",
]


class ConnectionsFile:
    def __init__(self, path_to_file) -> None:
        self.parser = configparser.ConfigParser()
        dsn_file = Path(path_to_file)

        cfg_content = dsn_file.read_text()
        self.parser.read_string(cfg_content)

    def get_default_connection_url(self):
        try:
            section = self.parser.items("default")
        except configparser.NoSectionError:
            return None

        url = URL.create(**dict(section))
        return str(url.render_as_string(hide_password=False))


def connection_str_from_dsn_section(section, config):
    """Return a SQLAlchemy connection string from a section in a DSN file

    Parameters
    ----------
    section : str
        The section name in the DSN file

    config : Config
        The config object, must have a dsn_filename attribute
    """
    parser = configparser.ConfigParser()
    dsn_file = Path(config.dsn_filename)

    try:
        cfg_content = dsn_file.read_text()
    except FileNotFoundError as e:
        raise exceptions.FileNotFoundError(
            f"%config SqlMagic.dsn_filename ({str(config.dsn_filename)!r}) not found."
            " Ensure the file exists or change the configuration: "
            "%config SqlMagic.dsn_filename = 'path/to/file.ini'"
        ) from e

    try:
        parser.read_string(cfg_content)
    except configparser.Error as e:
        raise exceptions.RuntimeError(
            "An error happened when loading "
            "your %config SqlMagic.dsn_filename "
            f"({config.dsn_filename!r})\n{type(e).__name__}: {e}"
        ) from e

    try:
        cfg = parser.items(section)
    except configparser.NoSectionError as e:
        raise exceptions.KeyError(
            f"The section {section!r} does not exist in the "
            f"connections file {config.dsn_filename!r}"
        ) from e

    cfg_dict = dict(cfg)

    try:
        url = URL.create(**cfg_dict)
    except TypeError as e:
        if "unexpected keyword argument" in str(e):
            raise exceptions.TypeError(
                f"%config SqlMagic.dsn_filename ({config.dsn_filename!r}) is invalid. "
                "It must only contain the following keys: drivername, username, "
                "password, host, port, database, query"
            ) from e
        else:
            raise

    return str(url.render_as_string(hide_password=False))


def _connection_string(arg, path_to_file):
    """
    Given a string, return a SQLAlchemy connection string if possible.

    Scenarios:

    - If the string is a valid URL, return it
    - If the string is a valid section in the DSN file return the connection string
    - Otherwise return an empty string

    Parameters
    ----------
    arg : str
        The string to parse

    path_to_file : str
        The path to the DSN file
    """
    # for environment variables
    arg = expandvars(arg)

    # if it's a URL, return it
    if "@" in arg or "://" in arg:
        return arg

    # if it's a section in the DSN file, return the connection string
    if arg.startswith("[") and arg.endswith("]"):
        section = arg.lstrip("[").rstrip("]")
        parser = configparser.ConfigParser()
        parser.read(path_to_file)
        cfg_dict = dict(parser.items(section))
        url = URL.create(**cfg_dict)
        url_ = str(url.render_as_string(hide_password=False))

        warnings.warn(
            "Starting connections with: %sql [section_name] is deprecated "
            "and will be removed in a future release. "
            "Please use: %sql --section section_name instead.",
            category=FutureWarning,
        )

        return url_

    return ""


def parse(arg, path_to_file):
    """Extract connection info and result variable from SQL

    Please don't add any more syntax requiring
    special parsing.
    Instead, add @arguments to SqlMagic.execute.

    We're grandfathering the
    connection string and `<<` operator in.

    Parameters
    ----------
    arg : str
        The string to parse

    path_to_file : str
        The path to the DSN file
    """
    result = {
        "connection": "",
        "sql": "",
        "result_var": None,
        "return_result_var": False,
    }

    pieces = arg.split(None, 1)
    if not pieces:
        return result

    result["connection"] = _connection_string(pieces[0], path_to_file)

    if result["connection"]:
        if len(pieces) == 1:
            return result
        arg = pieces[1]

    pointer = arg.find("<<")
    if pointer != -1:
        left = arg[:pointer].replace(" ", "").replace("\n", "")
        right = arg[pointer + 2 :].strip(" ")

        if "=" in left:
            result["result_var"] = left[:-1]
            result["return_result_var"] = True
        else:
            result["result_var"] = left

        result["sql"] = right
    else:
        result["sql"] = arg
    return result


def _option_strings_from_parser(parser):
    """Extracts the expected option strings (-a, --append, etc) from argparse parser

    Thanks Martijn Pieters
    https://stackoverflow.com/questions/28881456/how-can-i-list-all-registered-arguments-from-an-argumentparser-instance

    :param parser: [description]
    :type parser: IPython.core.magic_arguments.MagicArgumentParser
    """
    opts = [a.option_strings for a in parser._actions]
    return list(itertools.chain.from_iterable(opts))


def without_sql_comment(parser, line):
    """Strips -- comment from a line

    The argparser unfortunately expects -- to precede an option,
    but in SQL that delineates a comment.  So this removes comments
    so a line can safely be fed to the argparser.

    :param line: A line of SQL, possibly mixed with option strings
    :type line: str
    """

    args = _option_strings_from_parser(parser)
    result = itertools.takewhile(
        lambda word: (not word.startswith("--")) or (word in args),
        shlex.split(line, posix=False),
    )
    return " ".join(result)


def split_args_and_sql(line):
    """Separates line into args and sql query

    The argparser expects - to precede an argument, but postgreSQL
    and duckDB allow for -> and ->> to be used as JSON operators.
    This function splits the line into two - args and sql.
    This way we can only pass the args into the argparser, and
    add in the sql later.

    Parameters
    ----------
    line: str
        A line of SQL, preceded by option/argument strings

    Returns
    -------
    arg_line: str
        Portion of input line that contains only arguments

    sql_line: str
        Portion of input line that contains only SQL query/statements
    """
    arg_line, sql_line = line, ""

    # When queries include filenames, they may include SQL keywords
    #   ex. 'penguins_selected'.csv contains "select"
    # In these cases, splitting the query leads to parsing errors.
    # So we ignore any filenames by removing text between double quotes ""
    # and single quotes '' below.
    # Note: This won't affect the query because we are only modifying the
    # text we use to check for SQL commands. Any splitting is done
    # on the original line which includes filenames.
    line_no_filenames = re.sub(r"('.*')", "", line)  # 'file.csv' --> ''
    line_no_filenames = re.sub(r'(".*")', "", line_no_filenames)  # "file.csv" --> ""

    # Now that filenames are removed, check the line for any SQL commands
    # If any SQL commands are found in the line, we split the line into args and sql.
    #   Note: lines without SQL commands will not be split
    #       ex. %sql duckdb:// or %sqlplot boxplot --table data.csv
    if not any(cmd in line_no_filenames.lower() for cmd in SQL_COMMANDS):
        return arg_line, sql_line

    # Identify beginning of sql query using keywords
    split_idx = -1
    for token in line.split():
        if token.lower() in SQL_COMMANDS:
            # Found index at which to split line
            split_idx = line.find(token)
            break

    # Split line into args and sql, beginning at sql keyword
    if split_idx != -1:
        arg_line, sql_line = line[:split_idx], line[split_idx:]

    return arg_line, sql_line


def magic_args(magic_execute, line, cmd_from, allowed_duplicates=None):
    """
    Returns the parsed arguments from the line as parsed by magic_execute
    """
    allowed_duplicates = allowed_duplicates or []
    line = without_sql_comment(parser=magic_execute.parser, line=line)
    arg_line, sql_line = split_args_and_sql(line)

    args = shlex.split(arg_line, posix=False)

    if len(args) > 1:
        check_duplicate_arguments(magic_execute, cmd_from, args, allowed_duplicates)

    parsed = magic_execute.parser.parse_args(args)

    if sql_line:
        if parsed.line != "":
            parsed.line.extend(shlex.split(sql_line, posix=False))
        else:
            parsed.line = shlex.split(sql_line, posix=False)

    return parsed


def escape_string_literals_with_colon_prefix(query):
    """
    Given a query, replaces all occurrences of ':variable' with '\:variable' and
    ":variable" with "\:variable" so that the query can be passed to sqlalchemy.text
    without the literals being interpreted as bind parameters. Also calls
    escape_string_slicing_with_colon_prefix(). It doesn't replace
    the occurrences of :variable (without quotes)
    """  # noqa

    # Define the regular expression pattern for valid Python identifiers
    identifier_pattern = r"\b[a-zA-Z_][a-zA-Z0-9_]*\b"

    double_quoted_variable_pattern = r'(?<!\\)":(' + identifier_pattern + r')(?<!\\)"'

    # Define the regular expression pattern for matching ':variable' format
    single_quoted_variable_pattern = r"(?<!\\)':(" + identifier_pattern + r")(?<!\\)\'"

    # Replace ":variable" and ':variable' with "\:variable"
    query_quoted = re.sub(double_quoted_variable_pattern, r'"\\:\1"', query)
    query_quoted = re.sub(single_quoted_variable_pattern, r"'\\:\1'", query_quoted)

    double_found = re.findall(double_quoted_variable_pattern, query)
    single_found = re.findall(single_quoted_variable_pattern, query)

    # Escape occurrences of : for string slicing
    query_quoted, _ = escape_string_slicing_notation(query_quoted)

    return query_quoted, double_found + single_found


def escape_string_slicing_notation(query):
    """
    Given a query, replaces all occurrences of 'example'[x:y] with 'example'[x\:y].
    Escaping the colon using \ ensures correct string slicing behavior rather
    than being interpreted as a bind parameter.

    Parameters
    ----------
    query: str
        query to be parsed and cleaned
    """  # noqa
    identifier_pattern = r"\b[0-9_]*\b"

    # Define the regular expression pattern for matching [x:y]
    string_slicing_pattern = r"(?<!\\):(" + identifier_pattern + r")(?<!\\)\]"

    # Replace [x:y] with [x\:y]
    query_escaped = re.sub(string_slicing_pattern, r"\\:\1]", query)

    occurences_found = re.findall(string_slicing_pattern, query)

    return query_escaped, occurences_found


def find_named_parameters(input_string):
    # Define the regular expression pattern for valid Python identifiers
    identifier_pattern = r"\b[a-zA-Z_][a-zA-Z0-9_]*\b"

    # Define the regular expression pattern for matching :variable format
    variable_pattern = r'(?<!["\'])\:(' + identifier_pattern + ")"

    # Use findall to extract all matches of :variable from the input string
    matches = re.findall(variable_pattern, input_string)

    return matches



from sql import display
from sql import util
from sql.store import get_all_keys
from sql.exceptions import RuntimeError, TableNotFoundError


ORIGINAL_ERROR = "\nOriginal error message from DB driver:\n"
CTE_MSG = (
    "If using snippets, you may pass the --with argument explicitly.\n"
    "For more details please refer: "
    "https://jupysql.ploomber.io/en/latest/compose.html#with-argument"
)
POSTGRES_MSG = """\nLooks like you have run into some issues.
                Review our DB connection via URL strings guide:
                https://jupysql.ploomber.io/en/latest/connecting.html .
                 Using Ubuntu? Check out this guide: "
                https://help.ubuntu.com/community/PostgreSQL#fe_sendauth:_
                no_password_supplied\n"""


def _snippet_typo_error_message(query):
    """Function to generate message for possible
    snippets if snippet name in user query is a
    typo
    """
    if query:
        tables = util.extract_tables_from_query(query)
        for table in tables:
            suggestions = util.find_close_match(table, get_all_keys())
            err_message = f"There is no table with name {table!r}."
            if len(suggestions) > 0:
                # If snippet is found in suggestions, this snippet
                # must not be misspelled (a different table name is)
                # so we don't show this message.
                if table in suggestions:
                    continue
                suggestions_message = util.get_suggestions_message(suggestions)
                return f"{err_message}{suggestions_message}"
    return ""


def _detailed_message_with_error_type(error, query):
    """Function to generate descriptive error message.
    Currently it handles syntax error messages, table not found messages
    and password issue when connecting to postgres
    """
    original_error = str(error)
    syntax_error_substrings = [
        "syntax error",
        "error in your sql syntax",
        "incorrect syntax",
        "invalid sql",
        "syntax_error",
    ]
    not_found_substrings = [
        r"(\btable with name\b).+(\bdoes not exist\b)",
        r"(\btable\b).+(\bdoes not exist\b)",
        r"(\bobject\b).+(\bdoes not exist\b)",
        r"(\brelation\b).+(\bdoes not exist\b)",
        r"(\btable\b).+(\bdoesn't exist\b)",
        "not found",
        "could not find",
        "no such table",
    ]
    if util.if_substring_exists(original_error.lower(), syntax_error_substrings):
        return f"{CTE_MSG}\n\n{ORIGINAL_ERROR}{original_error}\n", RuntimeError
    elif util.if_substring_exists(original_error.lower(), not_found_substrings):
        typo_err_msg = _snippet_typo_error_message(query)
        if typo_err_msg:
            return (
                f"{CTE_MSG}\n\n{typo_err_msg}\n\n"
                f"{ORIGINAL_ERROR}{original_error}\n",
                TableNotFoundError,
            )
        else:
            return (
                f"{CTE_MSG}\n\n{ORIGINAL_ERROR}{original_error}\n",
                RuntimeError,
            )
    elif "fe_sendauth: no password supplied" in original_error:
        return f"{POSTGRES_MSG}\n{ORIGINAL_ERROR}{original_error}\n", RuntimeError
    return None, None


def _display_error_msg_with_trace(error, message):
    """Displays the detailed error message and prints
    original stack trace as well."""
    if message is not None:
        display.message(message)
    error.modify_exception = True
    raise error


def _raise_error(error, message, error_type):
    """Raise specific error from the detailed message. If detailed
    message is None reraise original error"""
    if message is not None:
        raise error_type(message) from error
    else:
        raise RuntimeError(str(error)) from error


def handle_exception(error, query=None, short_error=True):
    """
    This function is the entry point for detecting error type
    and handling it accordingly.
    """
    if util.is_sqlalchemy_error(error) or util.is_non_sqlalchemy_error(error):
        detailed_message, error_type = _detailed_message_with_error_type(error, query)
        if short_error:
            _raise_error(error, detailed_message, error_type)
        else:
            _display_error_msg_with_trace(error, detailed_message)
    else:
        raise error



import re
from jinja2 import Template


def load_file(file_path) -> str:
    """
    Returns the content of a file
    """
    with open(file_path, mode="r") as file:
        return file.read()


def load_js(*files) -> str:
    """
    Loads js files into HTML <script>

    Parameters
    ----------
    *files : str or a sequence of (file_path : str, parameters : dict)
        The JS files to load.
    """

    js = ""

    for file in files:
        if isinstance(file, str):
            js += load_file(file)
        else:
            path = file[0]
            template_params = file[1]
            js_template = load_file(path)
            js_template = Template(js_template)
            js += js_template.render(template_params)

    return f"""
    <script>{js}</script>
    """


def load_css(*files) -> str:
    """
    Loads css files into HTML <style>
    """
    css = ""

    for file in files:
        css += load_file(file)

    return f"""
    <style>{css}</style>
    """


def set_template_params(**kwargs):
    """
    Returns parameters in a dict format for Jinja2 template.

    We can use it when loading JS files with custom parameters.

    e.g.
    html_scripts = utils.load_js([path_to_file,
                                    set_template_params(
                                        param_one = 1,
                                        param_one = 2)
                                    ]
                                    )
    """
    return kwargs


def extract_function_by_name(source, function_name) -> str:
    """
    Return function str by name from string

    Parameters
    ----------
    source : str
        Text to extract JS function from

    function_name : str
        The name of the function to extract
    """
    pattern = (
        r"function\s+"
        + function_name
        + r"\s*\([^)]*\)\s*\{((?:[^{}]+|\{(?:[^{}]+|\{[^{}]*\})*\})*)\}"
    )
    match = re.search(pattern, source)
    if match:
        return match.group(0)
    else:
        return None



from traitlets import TraitError, TraitType
from sql import display
import warnings

VALUE_WARNING = (
    'Please use a valid option: "warn", "enabled", or "disabled". \n'
    "For more information, "
    "see the docs: "
    "https://jupysql.ploomber.io/en/latest/api/configuration.html#named-parameters"
)


class Parameters(TraitType):
    def __init__(self, **kwargs):
        super(Parameters, self).__init__(**kwargs)

    def validate(self, obj, value):
        if isinstance(value, bool):
            if value:
                warnings.warn(
                    "named_parameters: boolean values are now deprecated. "
                    f'Value {value} will be treated as "enabled". \n'
                    f"{VALUE_WARNING}",
                    FutureWarning,
                )
                return "enabled"
            else:
                warnings.warn(
                    "named_parameters: boolean values are now deprecated. "
                    f'Value {value} will be treated as "warn" (default). \n'
                    f"{VALUE_WARNING}",
                    FutureWarning,
                )
                return "warn"
        elif isinstance(value, str):
            if not value:
                display.message(
                    'named_parameters: Value "" will be treated as "warn" (default)'
                )
                return "warn"

            value = value.lower()
            if value not in ("warn", "enabled", "disabled"):
                raise TraitError(
                    f"{value} is not a valid option for named_parameters. "
                    f'Valid options are: "warn", "enabled", or "disabled".'
                )

            return value

        else:
            raise TraitError(
                f"{value} is not a valid option for named_parameters. "
                f'Valid options are: "warn", "enabled", or "disabled".'
            )



from sql import inspect
from sql.cmd.cmd_utils import CmdParser
from sql.util import expand_args, is_rendering_required


def profile(others, user_ns):
    """
    Implementation of `%sqlcmd profile`
    This function takes in a string containing command line arguments,
    parses them to extract the name of the table, the schema, and the output location.
    It then retrieves statistical information about the specified table and either
    returns the report or writes it to the specified location.
    It also uses the kernel namespace for expanding arguments declared as variables.


    Parameters
    ----------
    others : str,
        A string containing the command line arguments.

    user_ns : dict,
        User namespace of IPython kernel

    Returns
    -------
    report: PrettyTable
        statistics of the table
    """
    parser = CmdParser()
    parser.add_argument("-t", "--table", type=str, help="Table name", required=True)

    parser.add_argument("-s", "--schema", type=str, help="Schema name", required=False)

    parser.add_argument(
        "-o", "--output", type=str, help="Store report location", required=False
    )

    args = parser.parse_args(others)
    if is_rendering_required(" ".join(others)):
        expand_args(args, user_ns)

    report = inspect.get_table_statistics(schema=args.schema, name=args.table)

    if args.output:
        with open(args.output, "w") as f:
            f.write(report._repr_html_())

    return report



from jinja2 import Template
import math
import sql.connection
from sql.telemetry import telemetry
from sql.util import enclose_table_with_double_quotations


class facet:
    def __init__():
        pass

    def get_facet_values(self, table, column, with_):
        conn = sql.connection.ConnectionManager.current
        table = enclose_table_with_double_quotations(table, conn)
        template = Template(
            """
            SELECT
            distinct ({{column}})
            FROM {{table}}
            ORDER BY {{column}}
            """
        )
        query = template.render(table=table, column=column)

        values = conn.execute(query, with_).fetchall()
        # Added to make histogram more inclusive to NULLs
        # Filter out NULL values
        # If value[0] is NULL we skip it

        values = [value for value in values if value[0] is not None]
        n_plots = len(values)
        n_cols = len(values) if len(values) < 3 else 3
        n_rows = math.ceil(n_plots / n_cols)
        return values, n_rows, n_cols


class facet_wrap(facet):
    """
    Splits a plot into a matrix of panels

    Parameters
    ----------
    facet : str
        Column to groupby and plot on different panels.
    """

    @telemetry.log_call("facet-wrap-init")
    def __init__(self, facet: str, legend=True):
        self.facet = facet
        self.legend = legend

    def __add__(self, other):
        return other.__add__(other)



from ploomber_core.telemetry.telemetry import Telemetry

try:
    from importlib.metadata import version
except ModuleNotFoundError:
    from importlib_metadata import version


telemetry = Telemetry(
    api_key="phc_P9SpSeypyPwxrMdFn2edOOEooQioF2axppyEeDwtMSP",
    package_name="jupysql",
    version=version("jupysql"),
)



from pathlib import Path
from jinja2 import Template

from sqlalchemy.engine import Engine

from sql import parse, exceptions
from sql.store import store
from sql.connection import ConnectionManager, is_pep249_compliant, is_spark
from sql.util import validate_nonidentifier_connection


class SQLPlotCommand:
    def __init__(self, magic, line) -> None:
        self.args = parse.magic_args(
            magic.execute, line, "sqlplot", allowed_duplicates=["-w", "--with"]
        )


class SQLCommand:
    """
    Encapsulates the parsing logic (arguments, SQL code, connection string, etc.)

    """

    def __init__(self, magic, user_ns, line, cell) -> None:
        self._line = line
        self._cell = cell

        self.args = parse.magic_args(
            magic.execute,
            line,
            "sql",
            allowed_duplicates=["-w", "--with", "--append", "--interact"],
        )

        # self.args.line (everything that appears after %sql/%%sql in the first line)
        # is split in tokens (delimited by spaces), this checks if we have one arg
        one_arg = len(self.args.line) == 1

        # NOTE: this is only used to determine if what the user passed looks like a
        # connection, we can simplify it
        if len(self.args.line) > 0 and self.args.line[0] in user_ns:
            conn = user_ns[self.args.line[0]]

            is_dbapi_connection_ = is_pep249_compliant(conn)
        else:
            is_dbapi_connection_ = False

        if (
            one_arg
            and self.args.line[0] in user_ns
            and (
                isinstance(user_ns[self.args.line[0]], Engine)
                or is_dbapi_connection_
                or is_spark(user_ns[self.args.line[0]])
            )
        ):
            line_for_command = []
            add_conn = True
        else:
            line_for_command = self.args.line
            add_conn = False

        if one_arg and self.args.line[0] in ConnectionManager.connections:
            line_for_command = []
            add_alias = True
        else:
            add_alias = False

        self.command_text = " ".join(line_for_command) + "\n" + cell

        if self.args.file:
            try:
                file_contents = Path(self.args.file).read_text()
            except FileNotFoundError as e:
                raise exceptions.FileNotFoundError(str(e)) from e

            self.command_text = file_contents + "\n" + self.command_text

        self.parsed = parse.parse(self.command_text, magic.dsn_filename)

        self.parsed["sql_original"] = self.parsed["sql"] = self._var_expand(
            self.parsed["sql"], user_ns
        )

        if add_conn:
            self.parsed["connection"] = user_ns[self.args.line[0]]

        if add_alias:
            self.parsed["connection"] = self.args.line[0]

        if self.args.with_:
            self.args.with_ = [
                Template(item).render(user_ns) for item in self.args.with_
            ]
            final = store.render(self.parsed["sql"], with_=self.args.with_)
            self.parsed["sql"] = str(final)

        if (
            one_arg
            and self.sql
            and not (add_conn or add_alias)
            and not (self.args.persist_replace or self.args.persist or self.args.append)
        ):
            # Apply strip to ensure whitespaces/linebreaks aren't passed
            validate_nonidentifier_connection(self.sql.strip().split(" ")[0].strip())

    @property
    def sql(self):
        """
        Returns the SQL query to execute, without any other options or arguments
        """
        return self.parsed["sql"]

    @property
    def sql_original(self):
        """
        Returns the raw SQL query. Might be different from `sql` if using --with
        """
        return self.parsed["sql_original"]

    @property
    def connection(self):
        """Returns the connection string"""
        return self.parsed["connection"]

    @property
    def result_var(self):
        """Returns the result_var"""
        return self.parsed["result_var"]

    @property
    def return_result_var(self):
        """Returns the return_result_var"""
        return self.parsed["return_result_var"]

    def _var_expand(self, sql, user_ns):
        return Template(sql).render(user_ns)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(line={self._line!r}, cell={self._cell!r}) -> "
            f"({self.sql!r}, {self.sql_original!r})"
        )

    def set_sql_with(self, with_):
        """
        Sets the final rendered SQL query using the WITH clause

        Parameters
        ----------
        with_ : list
        list of all subqueries needed to render the query
        """
        final = store.render(self.parsed["sql"], with_)
        self.parsed["sql"] = str(final)



from sql import plot
from sql.ggplot.geom.geom import geom
from sql.telemetry import telemetry


class geom_boxplot(geom):
    """
    Boxplot
    """

    def __init__(self):
        pass

    @telemetry.log_call("ggplot-boxplot")
    def draw(self, gg, ax=None):
        plot.boxplot(
            table=gg.table,
            column=gg.mapping.x,
            conn=gg.conn,
            with_=gg.with_,
            ax=ax or gg.axs[0],
        )

        return gg



try:
    from pyspark.sql import DataFrame
    from pyspark.sql.connect.dataframe import DataFrame as CDataFrame
except ModuleNotFoundError:
    DataFrame = None
    CDataFrame = None

from sql import exceptions


def handle_spark_dataframe(dataframe, should_cache=False):
    """Execute a ResultSet sqlaproxy using pysark module."""
    if not DataFrame and not CDataFrame:
        raise exceptions.MissingPackageError("pysark not installed")

    return SparkResultProxy(dataframe, dataframe.columns, should_cache)


class SparkResultProxy(object):
    """A fake class that pretends to behave like the ResultProxy from
    SqlAlchemy.
    """

    dataframe = None

    def __init__(self, dataframe, headers, should_cache):
        self.dataframe = dataframe
        self.fetchall = dataframe.collect
        self.rowcount = dataframe.count()
        self.keys = lambda: headers
        self.cursor = SparkCursor(headers)
        self.returns_rows = True
        if should_cache:
            self.dataframe.cache()

    def fetchmany(self, size):
        return self.dataframe.take(size)

    def fetchone(self):
        return self.dataframe.head()

    def close(self):
        self.dataframe.unpersist()


class SparkCursor(object):
    """Class to extend to give SqlAlchemy Cursor like behaviour"""

    description = None

    def __init__(self, headers) -> None:
        self.description = headers



import prettytable


class CustomPrettyTable(prettytable.PrettyTable):
    def add_rows(self, data):
        for row in data:
            formatted_row = []
            for cell in row:
                if isinstance(cell, str) and cell.startswith("http"):
                    formatted_row.append("<a href={}>{}</a>".format(cell, cell))
                else:
                    formatted_row.append(cell)
            self.add_row(formatted_row)



from sql import plot
from sql.ggplot.geom.geom import geom
from sql.telemetry import telemetry


class geom_histogram(geom):
    """
    Histogram plot

    Parameters
    ----------
    bins: int
        Number of bins

    fill : str
        Create a stacked graph which is a combination of
        'x' and 'fill'

    cmap : str, default 'viridis
        Apply a color map to the stacked graph

    breaks : list
        Divide bins with custom intervals

    binwidth : int or float
        Width of each bin
    """

    def __init__(
        self, bins=None, fill=None, cmap=None, breaks=None, binwidth=None, **kwargs
    ):
        self.bins = bins
        self.fill = fill
        self.cmap = cmap
        self.breaks = breaks
        self.binwidth = binwidth
        super().__init__(**kwargs)

    @telemetry.log_call("ggplot-histogram")
    def draw(self, gg, ax=None, facet=None):
        plot.histogram(
            table=gg.table,
            column=gg.mapping.x,
            cmap=self.cmap,
            bins=self.bins,
            conn=gg.conn,
            with_=gg.with_,
            category=self.fill,
            color=gg.mapping.fill,
            edgecolor=gg.mapping.color,
            facet=facet,
            ax=ax or gg.axs[0],
            breaks=self.breaks,
            binwidth=self.binwidth,
        )
        return gg
